package ctbrec.ui.controls.range;

import javafx.beans.property.DoubleProperty;
import javafx.beans.property.SimpleDoubleProperty;
import javafx.geometry.Orientation;
import javafx.scene.control.Control;
import javafx.scene.control.Skin;

public class RangeSlider<T extends Number> extends Control  {

    private static final String DEFAULT_STYLE_CLASS = "rangeslider";

    private Range<T> range;
    private DoubleProperty low;
    private DoubleProperty high;
    private boolean showTickMarks = false;
    private boolean showTickLabels = false;
    private Orientation orientation = Orientation.HORIZONTAL;

    public RangeSlider(Range<T> range) {
        this.range = range;
        low = new SimpleDoubleProperty(getMinimum().doubleValue());
        high = new SimpleDoubleProperty(getMaximum().doubleValue());
        getStyleClass().setAll(DEFAULT_STYLE_CLASS);
    }

    @Override
    protected Skin<?> createDefaultSkin() {
        return new RangeSliderSkin(this, new RangeSliderBehavior<>(this));
    }

    @Override
    public String getUserAgentStylesheet() {
        return RangeSlider.class.getResource("rangeslider.css").toExternalForm();
    }

    public DoubleProperty getLow() {
        return low;
    }

    public void setLow(T newPosition) {
        low.set(newPosition.doubleValue());
    }

    public DoubleProperty getHigh() {
        return high;
    }

    public void setHigh(T newPosition) {
        this.high.set(newPosition.doubleValue());
    }

    public T getMinimum() {
        return range.getMinimum();
    }

    public T getMaximum() {
        return range.getMaximum();
    }

    public boolean isShowTickMarks() {
        return showTickMarks;
    }

    public void setShowTickMarks(boolean showTickMarks) {
        this.showTickMarks = showTickMarks;
    }

    public boolean isShowTickLabels() {
        return showTickLabels;
    }

    public void setShowTickLabels(boolean showTickLabels) {
        this.showTickLabels = showTickLabels;
    }

    public Orientation getOrientation() {
        return orientation;
    }

    public void setOrientation(Orientation orientation) {
        this.orientation = orientation;
    }

    public Range<T> getRange() {
        return range;
    }
}
