package ctbrec.ui.controls.range;

import java.util.List;

import com.sun.javafx.scene.control.behavior.BehaviorBase; // NOSONAR
import com.sun.javafx.scene.control.inputmap.InputMap; // NOSONAR

public class RangeSliderBehavior<T extends Number> extends BehaviorBase<RangeSlider<T>> {

    private RangeSlider<T> rangeSlider;

    public RangeSliderBehavior(RangeSlider<T> rangeSlider) {
        super(rangeSlider);
        this.rangeSlider = rangeSlider;
    }

    @Override
    public InputMap<RangeSlider<T>> getInputMap() {
        InputMap<RangeSlider<T>> inputMap = new InputMap<>(rangeSlider);
        return inputMap;
    }

    /**
     * @param position
     *            The mouse position on track with 0.0 being beginning of track and 1.0 being the end
     */
    public void lowThumbDragged(double position) {
        rangeSlider.setLow(getNewPosition(position));
    }

    /**
     * @param position
     *            The mouse position on track with 0.0 being beginning of track and 1.0 being the end
     */
    public void highThumbDragged(double position) {
        rangeSlider.setHigh(getNewPosition(position));
    }

    /**
     * Calculate the new position of the thumb given the clicked/dragged position
     *
     * @param position
     *            clicked position
     * @return new position
     */
    private T getNewPosition(double position) {
        List<T> ticks = rangeSlider.getRange().getTicks();
        double percentPerTick = 1d / (ticks.size() - 1);
        int index = (int) Math.round(position / percentPerTick);
        index = Math.min(ticks.size() - 1, Math.max(0, index));
        return ticks.get(index);
    }
}
